#!/usr/bin/env python3
import sys, os, os.path, argparse
import socket, struct, hashlib, math
import glob, datetime, random

# ATTENTION. This code is only for demonstration and learning only.
# It's not supposed to be used in serious/production job.
# Beware! It may have bugs and vulnerabilities.
# For serious job, use OpenSSL, LibreSSL, etc...

# ToyTLS version 1 (Oct-2023) supports:
# TLS 1.2
# TLS_RSA_WITH_NULL_SHA
# TLS_RSA_WITH_AES_128_CBC_SHA
# TLS_RSA_WITH_AES_256_CBC_SHA256

# ToyTLS version 2 (Oct-2023) also supports:
# TLS_DHE_RSA_WITH_AES_128_CBC_SHA

# https://pypi.org/project/hexdump/
# pip3 install hexdump
import hexdump

# https://cryptography.io/
# pip3 install cryptography
import cryptography.x509
import cryptography.hazmat.primitives.hashes
import cryptography.hazmat.primitives.ciphers
import cryptography.hazmat.backends
import cryptography.hazmat.primitives.hmac

# Constants:

TLS_RSA_WITH_NULL_SHA=(0, 2)
TLS_RSA_WITH_AES_128_CBC_SHA=(0, 0x2f)
TLS_RSA_WITH_AES_256_CBC_SHA256=(0, 0x3D)
# https://ciphersuite.info/cs/TLS_DHE_RSA_WITH_AES_128_CBC_SHA/
TLS_DHE_RSA_WITH_AES_128_CBC_SHA=(0, 0x33)
TLS_EMPTY_RENEGOTIATION_INFO_SCSV=(0, 0xff)

# ciphersuites supported in this version:
ciphersuites_supported=[
TLS_RSA_WITH_NULL_SHA,
TLS_RSA_WITH_AES_128_CBC_SHA,
TLS_RSA_WITH_AES_256_CBC_SHA256,
TLS_DHE_RSA_WITH_AES_128_CBC_SHA,
]

ciphersuite_to_str={
TLS_RSA_WITH_NULL_SHA             : "TLS_RSA_WITH_NULL_SHA",
TLS_RSA_WITH_AES_128_CBC_SHA      : "TLS_RSA_WITH_AES_128_CBC_SHA",
TLS_RSA_WITH_AES_256_CBC_SHA256   : "TLS_RSA_WITH_AES_256_CBC_SHA256",
TLS_DHE_RSA_WITH_AES_128_CBC_SHA  : "TLS_DHE_RSA_WITH_AES_128_CBC_SHA",
TLS_EMPTY_RENEGOTIATION_INFO_SCSV : "TLS_EMPTY_RENEGOTIATION_INFO_SCSV",
}

# from ssl3.h in openssl
SSL3_AL_WARNING=1
SSL3_AL_FATAL=2

SSL3_AD_CLOSE_NOTIFY=0
SSL3_AD_UNEXPECTED_MESSAGE=10#/* fatal */
SSL3_AD_BAD_RECORD_MAC=20#/* fatal */
decryption_failed=21
SSL3_AD_DECOMPRESSION_FAILURE=30#/* fatal */
SSL3_AD_HANDSHAKE_FAILURE=40#/* fatal */
SSL3_AD_NO_CERTIFICATE=41
SSL3_AD_BAD_CERTIFICATE=42
SSL3_AD_UNSUPPORTED_CERTIFICATE=43
SSL3_AD_CERTIFICATE_REVOKED=44
SSL3_AD_CERTIFICATE_EXPIRED=45
SSL3_AD_CERTIFICATE_UNKNOWN=46
SSL3_AD_ILLEGAL_PARAMETER=47#/* fatal */
decode_error=50

alert_type_s={
SSL3_AD_CLOSE_NOTIFY                : "SSL3_AD_CLOSE_NOTIFY",
SSL3_AD_UNEXPECTED_MESSAGE          : "SSL3_AD_UNEXPECTED_MESSAGE",
SSL3_AD_BAD_RECORD_MAC              : "SSL3_AD_BAD_RECORD_MAC",
decryption_failed                   : "decryption_failed",
SSL3_AD_DECOMPRESSION_FAILURE       : "SSL3_AD_DECOMPRESSION_FAILURE",
SSL3_AD_HANDSHAKE_FAILURE           : "SSL3_AD_HANDSHAKE_FAILURE",
SSL3_AD_NO_CERTIFICATE              : "SSL3_AD_NO_CERTIFICATE",
SSL3_AD_BAD_CERTIFICATE             : "SSL3_AD_BAD_CERTIFICATE",
SSL3_AD_UNSUPPORTED_CERTIFICATE     : "SSL3_AD_UNSUPPORTED_CERTIFICATE",
SSL3_AD_CERTIFICATE_REVOKED         : "SSL3_AD_CERTIFICATE_REVOKED",
SSL3_AD_CERTIFICATE_EXPIRED         : "SSL3_AD_CERTIFICATE_EXPIRED",
SSL3_AD_CERTIFICATE_UNKNOWN         : "SSL3_AD_CERTIFICATE_UNKNOWN",
SSL3_AD_ILLEGAL_PARAMETER           : "SSL3_AD_ILLEGAL_PARAMETER",
decode_error                        : "decode_error"
}

# content type. ssl3.h in openssl
SSL3_RT_CHANGE_CIPHER_SPEC=20
SSL3_RT_ALERT=21
SSL3_RT_HANDSHAKE=22
SSL3_RT_APPLICATION_DATA=23

def i2b(i):
    return struct.pack("<B", i)

# high byte + low byte. big-endian, as used in TLS
def i2w(i):
    return struct.pack(">H", i)

# sent as 03 00, 03 04, etc. big-endian.
SSL3_VERSION=0x0300
TLS1_VERSION=0x0301
TLS1_1_VERSION=0x0302
TLS1_2_VERSION=0x0303
TLS1_3_VERSION=0x0304
cur_TLS_ver=TLS1_2_VERSION
cur_TLS_ver_bin=i2w(cur_TLS_ver)

# handshake type. ssl3.h in openssl
SSL3_MT_CLIENT_HELLO=1
SSL3_MT_SERVER_HELLO=2
SSL3_MT_NEWSESSION_TICKET=4
SSL3_MT_CERTIFICATE=11
SSL3_MT_SERVER_KEY_EXCHANGE=12
SSL3_MT_SERVER_DONE=14
SSL3_MT_CLIENT_KEY_EXCHANGE=16
SSL3_MT_FINISHED=20

TLSEXT_TYPE_server_name=0
TLSEXT_TYPE_signature_algorithms=13
TLSEXT_TYPE_encrypt_then_mac=22
TLSEXT_TYPE_extended_master_secret=23
TLSEXT_TYPE_session_ticket=35
TLSEXT_TYPE_renegotiate=0xff01

TLSEXT_SIGALG_rsa_pkcs1_sha256=0x0401   # badssl.com, yurichev.com
TLSEXT_SIGALG_rsa_pkcs1_sha1=0x0201     # badssl.com supports this
#TLSEXT_SIGALG_rsa_pss_pss_sha256=0x0809 # yet untested

def inverse_dict(d):
    return {v: k for k, v in d.items()}

str_to_ciphersuite=inverse_dict(ciphersuite_to_str)

# ToyTLS service functions:

# prf* functions: stolen from https://github.com/python-tls/tls/blob/master/tls/_common/prf.py
# https://tools.ietf.org/html/rfc5246#section-5

def prf_p_hash(hash_algorithm, secret, seed, output_length):
    result = bytearray()
    i = 1
    while len(result) < output_length:
        h = cryptography.hazmat.primitives.hmac.HMAC(secret, hash_algorithm, cryptography.hazmat.backends.default_backend())
        h.update(prf_a(secret, hash_algorithm, i, seed))
        h.update(seed)
        result.extend(h.finalize())
        i += 1
    return bytes(result[:output_length])

def prf_a(secret, hash_algorithm, n, seed):
    if n == 0:
        return seed
    else:
        h = cryptography.hazmat.primitives.hmac.HMAC(secret, hash_algorithm, cryptography.hazmat.backends.default_backend())
        h.update(prf_a(secret, hash_algorithm, n - 1, seed))
        return h.finalize()

def prf(secret, label, seed, hash_algorithm, output_length):
    return prf_p_hash(hash_algorithm, secret, label + seed, output_length)

# https://techcommunity.microsoft.com/t5/iis-support-blog/ssl-tls-alert-protocol-and-the-alert-codes/ba-p/377132
# TLS 1.2 https://datatracker.ietf.org/doc/html/rfc5246#appendix-A.3
# TLS 1.3 https://datatracker.ietf.org/doc/html/rfc8446#appendix-B.2
def chk_alert(_type, buf):
    assert _type==SSL3_RT_ALERT
    level=buf[0]
    err=buf[1]
    if level==SSL3_AL_WARNING:
        s="warning"
    elif level==SSL3_AL_FATAL:
        s="fatal"
    else:
        assert False
    s=s+": "+alert_type_s[err]
    if level==SSL3_AL_FATAL:
        print ("Alert: "+ s)
        print ("Exiting")
        exit(0)
    return

# according to https://datatracker.ietf.org/doc/html/rfc5246#section-6.2.3.2
# for example, for "GET /\n" it returns 16-byte buf:
# 47 45 54 20 2F 0A 09 09  09 09 09 09 09 09 09 09  GET /...........
def add_CBC_padding(buf):
    new_buf_len=align2grain(len(buf), AES_block_size)
    pad_bytes_to_add=new_buf_len-len(buf)
    #print ("pad_bytes_to_add", pad_bytes_to_add)
    if pad_bytes_to_add==0:
        pad_bytes_to_add=AES_block_size
    return buf + i2b(pad_bytes_to_add-1)*pad_bytes_to_add

def remove_CBC_padding(buf):
    last_byte=buf[-1] # count of padding bytes added
    return buf[0:len(buf)-last_byte-1]

def calc_MAC(MAC_key, buf):
    if ciphersuite_after_handshake==TLS_RSA_WITH_NULL_SHA:
        hash_func=cryptography.hazmat.primitives.hashes.SHA1()
    elif ciphersuite_after_handshake==TLS_RSA_WITH_AES_128_CBC_SHA:
        hash_func=cryptography.hazmat.primitives.hashes.SHA1()
    elif ciphersuite_after_handshake==TLS_RSA_WITH_AES_256_CBC_SHA256:
        hash_func=cryptography.hazmat.primitives.hashes.SHA256()
    elif ciphersuite_after_handshake==TLS_DHE_RSA_WITH_AES_128_CBC_SHA:
        hash_func=cryptography.hazmat.primitives.hashes.SHA1()
    else:
        assert False

    h=cryptography.hazmat.primitives.hmac.HMAC(MAC_key, hash_func, cryptography.hazmat.backends.default_backend())
    h.update(buf)
    return h.finalize()

#from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
def encrypt_MAC_send(_type, buf):
    global cur_seq_n_to_serv
    if verbosity>=1:
        print (f"encrypt_MAC_send(). cur_seq_n_to_serv: {cur_seq_n_to_serv} _type {_type}")
    if verbosity>=2:
        print ("buf to send:")
        hexdump.hexdump(buf)
    if ciphersuite_after_handshake==TLS_RSA_WITH_NULL_SHA:
        # Only add MAC:
        header=i2b(_type)+cur_TLS_ver_bin+i2w(len(buf))

        seq=struct.pack(">Q", cur_seq_n_to_serv) # 8 bytes
        header_for_MAC=seq+i2b(_type)+cur_TLS_ver_bin+i2w(len(buf))
        verify_data=calc_MAC(MAC_key_from_client_to_serv, header_for_MAC+buf)
        if verbosity>=2:
            print("encrypt_MAC_send() verify_data:")
            hexdump.hexdump(verify_data)
        header=i2b(_type)+cur_TLS_ver_bin+i2w(len(buf+verify_data))

        s.send(header)
        s.send(buf)
        s.send(verify_data)

        cur_seq_n_to_serv=cur_seq_n_to_serv+1
        return

    # Encrypt and MAC and send:
    seq=struct.pack(">Q", cur_seq_n_to_serv) # 8 bytes

    if cipher_mode=="MTE":
        add_MAC_after_encryption=False
        add_MAC_before_encryption=True
    elif cipher_mode=="ETM":
        add_MAC_after_encryption=True
        add_MAC_before_encryption=False
    else:
        assert False

    IV=b"\x00"*AES_block_size

    cipher=cryptography.hazmat.primitives.ciphers.Cipher(cryptography.hazmat.primitives.ciphers.algorithms.AES(encr_key_from_client_to_serv), cryptography.hazmat.primitives.ciphers.modes.CBC(IV_from_client_to_serv))
    encryptor=cipher.encryptor()
    # according to https://datatracker.ietf.org/doc/html/rfc5246#section-6.2.3.2
    if add_MAC_before_encryption:
        header_for_MAC=seq+i2b(_type)+cur_TLS_ver_bin+i2w(len(buf))
        if verbosity>=2:
            print ("encrypt_MAC_send() header_for_MAC:")
            hexdump.hexdump(header_for_MAC)
        buf_MAC=calc_MAC(MAC_key_from_client_to_serv, header_for_MAC+buf)
        if verbosity>=2:
            print("encrypt_MAC_send() buf to encrypt (MtE):")
            hexdump.hexdump(IV+add_CBC_padding(buf+buf_MAC))
        encrypted=encryptor.update(IV+add_CBC_padding(buf+buf_MAC)) + encryptor.finalize()
    else:
        if verbosity>=2:
            print("encrypt_MAC_send() buf to encrypt (EtM):")
            hexdump.hexdump(IV+add_CBC_padding(buf))
        encrypted=encryptor.update(IV+add_CBC_padding(buf)) + encryptor.finalize()

    if add_MAC_after_encryption:
        header_for_MAC=seq+i2b(_type)+cur_TLS_ver_bin+i2w(len(encrypted))
        verify_data=calc_MAC(MAC_key_from_client_to_serv, header_for_MAC+encrypted)
        header=i2b(_type)+cur_TLS_ver_bin+i2w(len(encrypted+verify_data))
    else:
        header=i2b(_type)+cur_TLS_ver_bin+i2w(len(encrypted))
    if verbosity>=2:
        print("encrypt_MAC_send() sending header:")
        hexdump.hexdump(header)
    s.send(header)
    if verbosity>=2:
        print("encrypt_MAC_send() sending encrypted:")
        hexdump.hexdump(encrypted)
    s.send(encrypted)
    if add_MAC_after_encryption:
        if verbosity>=2:
            print("encrypt_MAC_send() sending verify_data:")
            hexdump.hexdump(verify_data)
        s.send(verify_data)
    cur_seq_n_to_serv=cur_seq_n_to_serv+1

"""
Yes, probably clanky.
Other ideas:
https://stackoverflow.com/questions/17508741/how-to-receive-certain-size-of-data-in-socket-programming
https://stackoverflow.com/questions/55825905/how-can-i-reliably-read-exactly-n-bytes-from-a-tcp-socket
"""
def recv_at_least(s, size):
    buf=b""
    left=size
    while left>0:
        got=s.recv(left)
        if len(got)==0:
            print ("Warning. recv_at_least(): got nothing. probably it's supposed client should close connection?")
            raise ValueError
        left=left-len(got)
        buf=buf+got
    return buf

def get_data_from_serv():
    if verbosity>=1:
        print ("get_data_from_serv() start")
    s.settimeout(20) # In seconds. Due to slow badssl.com :(
    buf=recv_at_least(s, 5) # header len
    #if len(buf)==0:
    #    print ("Warning. get_data_from_serv(): got nothing. probably it's supposed client should close connection?")
    #    raise ValueError
    if verbosity>=2:
        print ("get_data_from_serv() buf:")
        hexdump.hexdump(buf)
    _type=buf[0]
    if verbosity>=1:
        print ("type", _type)
    proto=(buf[1] << 8) | buf[2]
    proto_bin=buf[1:3]
    if verbosity>=1:
        print ("proto", hex(proto))
    size=buf[3]<<8 | buf[4]
    if verbosity>=1:
        print ("expected size", hex(size))
    buf=recv_at_least(s, size) # header len

    if verbosity>=2:
        print ("got buf:")
        hexdump.hexdump(buf)
    return _type, proto_bin, buf

def prepend_16_bit_len_to_buf(buf):
    return i2w(len(buf))+buf

# from http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-3.html (3.3)
def align2grain (i, grain):
    return ((i + grain-1) & ~(grain-1))

def unpack_ServerHello_extensions(buf):
    global cipher_mode, use_extended_master_secret
    if verbosity>=2:
        print ("unpack_ServerHello_extensions")
        hexdump.hexdump(buf)
    idx=0
    length=(buf[idx]<<8) | buf[idx+1]
    idx=idx+2
    while idx<len(buf):
        if buf[idx]==0xff and buf[idx+1]==1:
            if verbosity>=1:
                print ("extension from server: renegotiation_info")
            idx=idx+5
        elif buf[idx]==0 and buf[idx+1]==0:
            if verbosity>=1:
                print ("extension from server: server_name")
            idx=idx+4
        elif buf[idx]==0 and buf[idx+1]==0x23:
            if verbosity>=1:
                print ("extension from server: session_ticket")
            idx=idx+4
        elif buf[idx]==0 and buf[idx+1]==0x16:
            if verbosity>=1:
                print ("extension from server: encrypt_then_mac")
            idx=idx+4
            # server agreed to use EtM
            cipher_mode="ETM"
        elif buf[idx]==0 and buf[idx+1]==0x17:
            if verbosity>=1:
                print ("extension from server: extended_master_secret")
            idx=idx+4
            # server agreed to use extended_master_secret
            use_extended_master_secret=True
        else:
            if verbosity>=1:
                print ("warning: unknown (yet?) extension", buf[idx], buf[idx+1])
            idx=idx+4
    assert idx==len(buf)
    return length+2 # 2 bytes were used for encoding 16-bit length

# this packet is hashed in ssl3_finish_mac()
def unpack_SSL3_MT_SERVER_HELLO(buf):
    if verbosity>=1:
        print ("unpack_SSL3_MT_SERVER_HELLO")
    global server_random
    global MAC_size
    global AES_block_size
    global AES_key_size
    global ciphersuite_after_handshake

    if verbosity>=2:
        hexdump.hexdump(buf)
    idx=0
    handshake_type=buf[idx]
    assert handshake_type==SSL3_MT_SERVER_HELLO
    idx=idx+1
    length=(buf[idx]<<16) | (buf[idx+1]<<8) | buf[idx+2]
    idx=idx+3
    version=(buf[idx]<<8) | buf[idx+1]
    idx=idx+2
    server_random=buf[idx:idx+0x20]
    if verbosity>=2:
        print ("server random:")
        hexdump.hexdump(server_random)
    idx=idx+0x20
    session_id_len=buf[idx]
    idx=idx+1
    if session_id_len!=0:
        print ("Warning, session_id is present, but unhandled")
    idx=idx+session_id_len
    ciphersuite_H=buf[idx]
    ciphersuite_L=buf[idx+1]
    idx=idx+2
    if verbosity>=1:
        print ("Server picked ciphersuite:", ciphersuite_to_str[(ciphersuite_H, ciphersuite_L)])
    if (ciphersuite_H, ciphersuite_L)==(0,2):
        ciphersuite_after_handshake=TLS_RSA_WITH_NULL_SHA
        MAC_size=20
        AES_block_size=16
        AES_key_size=16
    elif (ciphersuite_H, ciphersuite_L)==(0, 0x2f):
        ciphersuite_after_handshake=TLS_RSA_WITH_AES_128_CBC_SHA
        MAC_size=20
        AES_block_size=16
        AES_key_size=16
    elif (ciphersuite_H, ciphersuite_L)==(0, 0x3d):
        ciphersuite_after_handshake=TLS_RSA_WITH_AES_256_CBC_SHA256
        MAC_size=32
        AES_block_size=16
        AES_key_size=32
    elif (ciphersuite_H, ciphersuite_L)==(0, 0x33):
        ciphersuite_after_handshake=TLS_DHE_RSA_WITH_AES_128_CBC_SHA
        MAC_size=20
        AES_block_size=16
        AES_key_size=16
    else:
        print ("Error. Server offers ciphersuite we don't support.")
        exit(0)

    compression_method=buf[idx] 
    idx=idx+1
    #print ("compression_method", compression_method, hex(compression_method))
    assert compression_method==0 # no other handled
    serv_ext_len=unpack_ServerHello_extensions(buf[idx:])
    idx=idx+serv_ext_len
    assert idx==len(buf) # Make sure all buffer is processed

# This is a 'high-level' method of checking
def chk_cert_1(cert, issuer):
    # Verify by specialized function:
    try:
        cert.verify_directly_issued_by(issuer)
    except ValueError:
        return False
    return True

# This is a 'low-level' method of checking
def chk_cert_2(cert, issuer):
    # Verify RSA signature
    try:
        issuer.public_key().verify(
            cert.signature,
            cert.tbs_certificate_bytes,
            cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15(),
            cert.signature_hash_algorithm,
        )
    except cryptography.exceptions.InvalidSignature:
        return False
    except TypeError: # In case if issuer has EC pub key. We don't handle EC anyway.
        return False
    return True

def chk_cert(cert, issuer):
    not_valid_before=cert.not_valid_before
    not_valid_after=cert.not_valid_after

    current_datetime=datetime.datetime.now()
    if current_datetime<=not_valid_before or current_datetime>=not_valid_after:
        print ("Error. Certificate is expired")
        print ("cert not valid before", not_valid_before)
        print ("cert not valid after", not_valid_after)
        exit(0)

    # There are at least two ways of cert checking here
    # I use two here intentionally
    # To show that the task can be solved in two ways
    x=chk_cert_1(cert, issuer)
    y=chk_cert_2(cert, issuer)
    # Both methods must report identical result:
    assert x==y
    return x

def cert_self_signed(cert):
    issuer=cert.issuer
    subject=cert.subject

    return issuer==subject

def chk_cert_chain(certs_from_serv):
    all_certs=[]

    if os.path.isdir(local_cert_path)==False:
        print (f"Error. There is no directory with local store: {local_cert_path}.")
        print ("Use -local_cert_path option to correct this.")
        print ("Or use the 'local_cert_store_from_Ubuntu_22' directory.")
        print ("Can't check cert.")
        return

    for x in os.listdir(local_cert_path+"/"):
        if os.path.isdir(local_cert_path+"/"+x):
            continue
        f=open(local_cert_path+"/"+x,"rb")
        certs=cryptography.x509.load_pem_x509_certificates(f.read())
        all_certs=all_certs+certs
        f.close()
    if verbosity>=1:
        print (f"Certs imported from {local_cert_path}: {len(all_certs)}")

    for serv_cert in certs_from_serv:
        for issuer in all_certs:
            if chk_cert(serv_cert, issuer):
                if verbosity>=1:
                    print ("Cert checked. Found the issuer in local store.")
                    print ("Cert from serv:", serv_cert)
                    print ("Issuer from local store:", issuer)
                    #exit(0)
                return
    print ("ATTENTION!")
    print ("I can't check (any) cert from serv.")
    print ("IOW, I can't find an issuer in local store.")
    if len(certs_from_serv)==1 and cert_self_signed(certs_from_serv[0]):
        print ("Also, the sole cert is self-signed.")
    print ("This is a sign of misconfigured webserver or MITM attack.")
    #print ("Stopping.")
    #exit(0)

# this packet is hashed in ssl3_finish_mac()
def unpack_SSL3_MT_CERTIFICATE(buf):
    if verbosity>=1:
        print ("unpack_SSL3_MT_CERTIFICATE")
    global RSA_pub_key

    if verbosity>=2:
        hexdump.hexdump(buf)
    idx=0
    handshake_type=buf[idx]
    assert handshake_type==SSL3_MT_CERTIFICATE
    idx=idx+1
    length=(buf[idx]<<16) | (buf[idx+1]<<8) | buf[idx+2]
    if verbosity>=1:
        print ("length", length)
    idx=idx+3
    certs_length=(buf[idx]<<16) | (buf[idx+1]<<8) | buf[idx+2]
    if verbosity>=1:
        print ("certs_length", certs_length)
    idx=idx+3
    # First cert must have CN=host
    # Last cert is supposed to be installed on client's OS/browser
    certs_from_serv=[]
    while idx<len(buf):
        cert_length=(buf[idx]<<16) | (buf[idx+1]<<8) | buf[idx+2]
        if verbosity>=1:
            print ("cert_length", cert_length)
        idx=idx+3
        certs=buf[idx:idx+cert_length]
        x=cryptography.x509.load_der_x509_certificate(certs)
        if verbosity>=1:
            print ("Got cert", x)
            print ("Issuer", x.issuer)
        if cert_self_signed(x):
            if verbosity>=1:
                print ("This is a self-signed cert!")
        if len(certs_from_serv)==0:
            # got first cert
            tmp=x.subject.get_attributes_for_oid(cryptography.x509.NameOID.COMMON_NAME)
            if len(tmp)>0:
                CN=tmp[0].value
                if CN[0]=="*":
                    # https://en.wikipedia.org/wiki/Wildcard_certificate
                    print (f"Warning: mask in wildcard certs is not handled yet: {CN}, but cert is checked anyway")
                elif CN!=HOST:
                    print ("Error.")
                    print ("Certificate's CN doesn't correspond to host you connect")
                    print ("CN from cert:", CN)
                    print ("Host you're trying to connect:", HOST)
                    print ("Often, this is just misconfigured webserver")
                    exit(0)

        if len(certs_from_serv)>0:
            # check against previous
            prev_cert=certs_from_serv[-1]
            cur_cert=x

            if prev_cert==cur_cert:
                print ("Warning, two identical certs in (full)chain")
                print ("prev_cert", prev_cert)
                print ("cur_cert", cur_cert)
                print ("This is a sign of misconfigured webserver.")

            # https://datatracker.ietf.org/doc/html/rfc5246#section-7.4.2
            if prev_cert!=cur_cert and chk_cert(prev_cert, cur_cert)==False:
                print ("Warning. accoring to TLS standard, prev_cert in (full)chain must be issued by cur_cert.")
                print ("But this is not the case.")
                print ("prev_cert", prev_cert)
                print ("cur_cert", cur_cert)

        certs_from_serv.append(x)
        pub_key=x.public_key()
        if len(certs_from_serv)==1:
            RSA_pub_key=pub_key # set it to the first cert we've seen. we'll use it
            if verbosity>=2:
                print ("RSA pub key in the (first) cert:")
                print ("e", RSA_pub_key.public_numbers().e)
                print ("n", RSA_pub_key.public_numbers().n)
                print ("log_2(n)", math.log(RSA_pub_key.public_numbers().n, 2))
        idx=idx+cert_length

    assert idx==len(buf) # Make sure this is the packet end

    chk_cert_chain(certs_from_serv)

# this packet is hashed in ssl3_finish_mac() (openssl)
def unpack_SSL3_MT_SERVER_DONE(buf):
    if verbosity>=1:
        print ("unpack_SSL3_MT_SERVER_DONE")
    if verbosity>=2:
        hexdump.hexdump(buf)
    assert buf[0]==SSL3_MT_SERVER_DONE

# cipher suites length + cipher suites
def create_cipher_suites():
    global MAC_size, AES_block_size, AES_key_size
    tmp=b""
    # FIXME: use ciphersuites_supported[] here
    if the_only_ciphersuite_client_wants=="any":
        tmp=tmp+struct.pack("<BB", TLS_RSA_WITH_AES_128_CBC_SHA[0], TLS_RSA_WITH_AES_128_CBC_SHA[1])
        tmp=tmp+struct.pack("<BB", TLS_RSA_WITH_AES_256_CBC_SHA256[0], TLS_RSA_WITH_AES_256_CBC_SHA256[1])
        tmp=tmp+struct.pack("<BB", TLS_RSA_WITH_NULL_SHA[0], TLS_RSA_WITH_NULL_SHA[1])
        tmp=tmp+struct.pack("<BB", TLS_EMPTY_RENEGOTIATION_INFO_SCSV[0], TLS_EMPTY_RENEGOTIATION_INFO_SCSV[1])
    else:
        x=str_to_ciphersuite[the_only_ciphersuite_client_wants]
        tmp=tmp+struct.pack("<BB", x[0], x[1])
    return prepend_16_bit_len_to_buf(tmp)

def create_list_of_extension_signature_algorithms():
    tmp=b""
    tmp=tmp+struct.pack(">H", TLSEXT_SIGALG_rsa_pkcs1_sha256)
    tmp=tmp+struct.pack(">H", TLSEXT_SIGALG_rsa_pkcs1_sha1)
    #tmp=tmp+struct.pack(">H", TLSEXT_SIGALG_rsa_pss_pss_sha256)
    tmp2=prepend_16_bit_len_to_buf(tmp) # signature hash algorithms length
    extension_signature_algorithms=b""
    extension_signature_algorithms=extension_signature_algorithms+struct.pack(">H", TLSEXT_TYPE_signature_algorithms)
    extension_signature_algorithms=extension_signature_algorithms+prepend_16_bit_len_to_buf(tmp2) # length
    return extension_signature_algorithms

def create_extensions():
    # extensions
    # most often: 16-bit - ext. code + 16-bit - length
    tmp=b""
    # server_name
    tmp=tmp+b"\x00\x00"+prepend_16_bit_len_to_buf(prepend_16_bit_len_to_buf(b"\x00"+prepend_16_bit_len_to_buf(bytes(HOST, "utf-8"))))
    tmp=tmp+struct.pack(">HH", TLSEXT_TYPE_session_ticket, 0)
    if DISABLE_ETM==False:
        tmp=tmp+struct.pack(">HH", TLSEXT_TYPE_encrypt_then_mac, 0)
    if TRY_EXTENDED_MASTER_SECRET:
        tmp=tmp+struct.pack(">HH", TLSEXT_TYPE_extended_master_secret, 0)
    return prepend_16_bit_len_to_buf(tmp+create_list_of_extension_signature_algorithms())

def make_hashes_for_client():
    m=hashlib.sha256()
    if verbosity>=2:
        print ("ClientHelloToHash to be hashed for 'client finished':")
        hexdump.hexdump(ClientHelloToHash)
    m.update(ClientHelloToHash)
    if verbosity>=2:
        print ("MT_SERVER_HELLO_pkt to be hashed for 'client finished':")
        hexdump.hexdump(MT_SERVER_HELLO_pkt)
    m.update(MT_SERVER_HELLO_pkt)
    if verbosity>=2:
        print ("MT_CERTIFICATE_pkt to be hashed for 'client finished':")
        hexdump.hexdump(MT_CERTIFICATE_pkt)
    m.update(MT_CERTIFICATE_pkt)
    if MT_SERVER_KEY_EXCHANGE_pkt!=None:
        if verbosity>=2:
            print ("MT_SERVER_KEY_EXCHANGE_pkt to be hashed for 'client finished':")
            hexdump.hexdump(MT_SERVER_KEY_EXCHANGE_pkt)
        m.update(MT_SERVER_KEY_EXCHANGE_pkt)
    if verbosity>=2:
        print ("MT_SERVER_DONE_pkt to be hashed for 'client finished':")
        hexdump.hexdump(MT_SERVER_DONE_pkt)
    m.update(MT_SERVER_DONE_pkt)
    if verbosity>=2:
        print ("client_key_exchange to be hashed for 'client finished':")
        hexdump.hexdump(client_key_exchange)
    m.update(client_key_exchange)
    return m.digest(), m.hexdigest()

def generate_key_block():
    global MAC_key_from_client_to_serv
    global MAC_key_from_serv_to_client
    global encr_key_from_client_to_serv
    global encr_key_from_serv_to_client
    global IV_from_client_to_serv
    global IV_from_serv_to_client

    key_block_size=MAC_size*2 + AES_key_size*2 + AES_block_size*2 # two MAC keys, two AES keys, two IVs for AES

    # tls1_generate_key_block() in openssl
    # http://www.faqs.org/rfcs/rfc5246.html 6.3
    key_block=prf(master_secret, b"key expansion", server_random+ClientHelloRandom, cryptography.hazmat.primitives.hashes.SHA256(), key_block_size)
    if verbosity>=2:
        print ("key_block:")
        hexdump.hexdump (key_block)

    idx=0
    MAC_key_from_client_to_serv=key_block[idx:idx+MAC_size]
    idx=idx+MAC_size
    if verbosity>=2:
        print ("MAC_key_from_client_to_serv:")
        hexdump.hexdump (MAC_key_from_client_to_serv)

    MAC_key_from_serv_to_client=key_block[idx:idx+MAC_size]
    if verbosity>=2:
        print ("MAC_key_from_serv_to_client:")
        hexdump.hexdump (MAC_key_from_serv_to_client)
    idx=idx+MAC_size

    encr_key_from_client_to_serv=key_block[idx:idx+AES_key_size]
    if verbosity>=2:
        print ("encr_key_from_client_to_serv:")
        hexdump.hexdump(encr_key_from_client_to_serv)
    idx=idx+AES_key_size

    encr_key_from_serv_to_client=key_block[idx:idx+AES_key_size]
    if verbosity>=2:
        print ("encr_key_from_serv_to_client")
        hexdump.hexdump(encr_key_from_serv_to_client)
    idx=idx+AES_key_size

    IV_from_client_to_serv=key_block[idx:idx+AES_block_size]
    if verbosity>=2:
        print ("IV_from_client_to_serv:")
        hexdump.hexdump(IV_from_client_to_serv)
    idx=idx+AES_block_size

    IV_from_serv_to_client=key_block[idx:idx+AES_block_size]
    if verbosity>=2:
        print ("IV_from_serv_to_client")
        hexdump.hexdump(IV_from_serv_to_client)
    idx=idx+AES_block_size

def remove_MAC_and_chk(header_for_MAC, buf):
    MAC=buf[-MAC_size:]
    if verbosity>=2:
        print ("remove_MAC_and_chk(): MAC:")
        hexdump.hexdump(MAC)

    body=buf[0:len(buf)-MAC_size]
    if verbosity>=1:
        print (f"remove_MAC_and_chk(): {len(body)}")

    calculated_MAC=calc_MAC(MAC_key_from_serv_to_client, header_for_MAC+body)
    if verbosity>=2:
        print("remove_MAC_and_chk(): calculated (on our side) MAC:")
        hexdump.hexdump(calculated_MAC)

    if MAC==calculated_MAC:
        if verbosity>=1:
             print ("remove_MAC_and_chk(): MAC is correct")
        return body, MAC
    else:
        print ("remove_MAC_and_chk(): MAC is incorrect. exiting.")
        exit(0)

def chk_MAC_and_decrypt(_type, serv_TLS_ver_bin, buf):
    global cur_seq_n_to_clnt
    if verbosity>=1:
        print (f"chk_MAC_and_decrypt(): start. cur_seq_n_to_clnt: {cur_seq_n_to_clnt}")
    seq=struct.pack(">Q", cur_seq_n_to_clnt) # 8 bytes

    # NULL cipher
    if ciphersuite_after_handshake==TLS_RSA_WITH_NULL_SHA:
        header_for_MAC=seq+i2b(_type)+serv_TLS_ver_bin+i2w(len(buf)-MAC_size)
        body, MAC=remove_MAC_and_chk(header_for_MAC, buf)

        if verbosity>=2:
            print ("chk_MAC_and_decrypt(): buf with MAC removed:")
            hexdump.hexdump(body)
        cur_seq_n_to_clnt=cur_seq_n_to_clnt+1
        return body

    # ETM mode
    # Used in newer TLS versions
    # https://datatracker.ietf.org/doc/html/rfc7366
    if cipher_mode=="ETM":
        header_for_MAC=seq+i2b(_type)+serv_TLS_ver_bin+i2w(len(buf)-MAC_size)
        buf_to_decrypt, MAC=remove_MAC_and_chk(header_for_MAC, buf)

        if verbosity>=1:
            print (f"chk_MAC_and_decrypt(): len(buf_to_decrypt) {len(buf_to_decrypt)}")
        cipher=cryptography.hazmat.primitives.ciphers.Cipher(cryptography.hazmat.primitives.ciphers.algorithms.AES(encr_key_from_serv_to_client), cryptography.hazmat.primitives.ciphers.modes.CBC(IV_from_serv_to_client))
        decryptor=cipher.decryptor()
        decrypted=decryptor.update(buf_to_decrypt) + decryptor.finalize()
        decrypted_stripped=remove_CBC_padding(decrypted[AES_block_size:])
        if verbosity>=2:
            print ("chk_MAC_and_decrypt(): decrypted (with IV and padding removed)")
            hexdump.hexdump(decrypted_stripped) # also skip heading IV
        cur_seq_n_to_clnt=cur_seq_n_to_clnt+1
        return decrypted_stripped

    # MTE mode
    # Used in older TLS versions
    # https://datatracker.ietf.org/doc/html/rfc5246#section-6.2.3.1
    # https://datatracker.ietf.org/doc/html/rfc5246#section-6.2.3.2
    if cipher_mode=="MTE":
        if verbosity>=1:
            print (f"chk_MAC_and_decrypt(): len(buf) {len(buf)}")
        cipher=cryptography.hazmat.primitives.ciphers.Cipher(cryptography.hazmat.primitives.ciphers.algorithms.AES(encr_key_from_serv_to_client), cryptography.hazmat.primitives.ciphers.modes.CBC(IV_from_serv_to_client))
        decryptor=cipher.decryptor()
        decrypted=decryptor.update(buf) + decryptor.finalize()
        if verbosity>=2:
            print ("chk_MAC_and_decrypt(): decrypted, as is:")
            hexdump.hexdump(decrypted)
        tmp=remove_CBC_padding(decrypted)
        IV=tmp[0:AES_block_size]
        tmp=tmp[AES_block_size:] # remove IV
        MAC=tmp[-MAC_size:]
        if verbosity>=2:
            print ("chk_MAC_and_decrypt(): MAC:")
            hexdump.hexdump(MAC)
        tmp=tmp[0:len(tmp)-MAC_size]
        if verbosity>=2:
            print ("chk_MAC_and_decrypt(): with IV and MAC and padding removed:")
            hexdump.hexdump(tmp)

        header_for_MAC=seq+i2b(_type)+serv_TLS_ver_bin+i2w(len(tmp))
        if verbosity>=2:
            print ("header_for_MAC:")
            hexdump.hexdump(header_for_MAC)
        buf_MAC=calc_MAC(MAC_key_from_serv_to_client, header_for_MAC+tmp)
        if verbosity>=2:
            print ("buf_MAC:")
            hexdump.hexdump(buf_MAC)
        if MAC==buf_MAC:
            if verbosity>=1:
                print ("chk_MAC_and_decrypt(): MAC is correct")
        else:
            print ("chk_MAC_and_decrypt(): MAC is incorrect. stopping.")
            exit(0)

        cur_seq_n_to_clnt=cur_seq_n_to_clnt+1
        return tmp

def make_hashes_for_serv():
    m=hashlib.sha256()

    m.update(ClientHelloToHash)
    if verbosity>=2:
        print ("ClientHelloToHash:")
        hexdump.hexdump(ClientHelloToHash)

    m.update(MT_SERVER_HELLO_pkt)
    if verbosity>=2:
        print ("MT_SERVER_HELLO_pkt:")
        hexdump.hexdump (MT_SERVER_HELLO_pkt)

    # cert
    m.update(MT_CERTIFICATE_pkt)
    if verbosity>=2:
        print ("MT_CERTIFICATE_pkt:")
        hexdump.hexdump (MT_CERTIFICATE_pkt)

    if MT_SERVER_KEY_EXCHANGE_pkt!=None:
        if verbosity>=2:
            print ("MT_SERVER_KEY_EXCHANGE_pkt to be hashed for 'client finished':")
            hexdump.hexdump(MT_SERVER_KEY_EXCHANGE_pkt)
        m.update(MT_SERVER_KEY_EXCHANGE_pkt)

    # server done, 4 bytes
    m.update(MT_SERVER_DONE_pkt)
    if verbosity>=2:
        print ("MT_SERVER_DONE_pkt:")
        hexdump.hexdump (MT_SERVER_DONE_pkt)

    m.update(client_key_exchange)
    if verbosity>=2:
        print ("client_key_exchange:")
        hexdump.hexdump(client_key_exchange)

    # 14 00... 16 bytes
    m.update(client_finished_pkt)
    if verbosity>=2:
        print ("client_finished_pkt:")
        hexdump.hexdump (client_finished_pkt)

    if MT_NEWSESSION_TICKET_pkt!=None:
        m.update(MT_NEWSESSION_TICKET_pkt) # new session ticket
        if verbosity>=2:
            print ("MT_NEWSESSION_TICKET_pkt:") # 04 00...
            hexdump.hexdump (MT_NEWSESSION_TICKET_pkt)
    return m.digest(), m.hexdigest()

def chk_server_finished(buf):
    global master_secret, handshake_hashes_clnt_bin, handshake_hashes_from_serv_bin

    if verbosity>=1:
        print ("chk_server_finished() start")

    handshake_hashes_from_serv_bin, handshake_hashes_from_serv_hex=make_hashes_for_serv()
    if verbosity>=2:
        print (f"handshake_hashes_from_serv: {handshake_hashes_from_serv_hex}")

    if verbosity>=2:
        print ("chk_server_finished(), buf:")
        hexdump.hexdump(buf)
    assert len(buf)==12
    if verbosity>=2:
        print ("master_secret:")
        hexdump.hexdump(master_secret)
        print ("handshake_hashes_from_serv_bin:")
        hexdump.hexdump(handshake_hashes_from_serv_bin)
    verify_data=prf(master_secret, b"server finished", handshake_hashes_from_serv_bin, cryptography.hazmat.primitives.hashes.SHA256(), 12)
    if verbosity>=2:
        print ("chk_server_finished(), verify_data:")
        hexdump.hexdump(verify_data)
    assert buf==verify_data # Make sure it's correct

def generate_master_secret():
    if use_extended_master_secret:
        # tls1_generate_master_secret() in openssl
        # https://www.ietf.org/rfc/rfc7627.html#section-4
        master_secret=prf(premaster_secret, b"extended master secret", handshake_hashes_clnt_bin, cryptography.hazmat.primitives.hashes.SHA256(), 48)
    else:
        # https://datatracker.ietf.org/doc/html/rfc5246#section-8.1
        master_secret=prf(premaster_secret, b"master secret", ClientHelloRandom+server_random, cryptography.hazmat.primitives.hashes.SHA256(), 48)
    if verbosity>=2:
        print ("master_secret:")
        hexdump.hexdump(master_secret)
    return master_secret

def generate_premaster_and_client_key_exchange():
    global dh_Ys
    if ciphersuite_after_handshake==TLS_DHE_RSA_WITH_AES_128_CBC_SHA:
        priv_rand=random.randint(1, dh_p-1)
        dh_Yc=pow(dh_g, priv_rand, dh_p) # to be sent to server
        shared_secret=pow(dh_Ys, priv_rand, dh_p)
        if verbosity>=1:
            print (f"generate_premaster_and_client_key_exchange() shared_secret {hex(shared_secret)}")
            print (f"generate_premaster_and_client_key_exchange() dh_Yc {hex(dh_Yc)}")
        to_send=prepend_16_bit_len_to_buf(dh_Yc.to_bytes(256, byteorder='big'))
        premaster_secret=shared_secret.to_bytes(256, byteorder='big')
        client_key_exchange=i2b(SSL3_MT_CLIENT_KEY_EXCHANGE)+b"\x00"+prepend_16_bit_len_to_buf(to_send)
        return premaster_secret, client_key_exchange

    # TLS_RSA_* ciphersuites

    # https://www.rfc-editor.org/rfc/rfc5246#section-7.4.7.1
    #premaster_secret=cur_TLS_ver_bin+(b"\x22"*(48-2)) # some determinism for debugging...
    premaster_secret=cur_TLS_ver_bin+os.urandom(48-2)

    encrypted_premaster_secret=RSA_pub_key.encrypt(premaster_secret, cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15())
    if verbosity>=2:
        print ("encrypted_premaster_secret")
        hexdump.hexdump(encrypted_premaster_secret)
    RSA_encrypted_premaster_secret=prepend_16_bit_len_to_buf(encrypted_premaster_secret)
    client_key_exchange=i2b(SSL3_MT_CLIENT_KEY_EXCHANGE)+b"\x00"+prepend_16_bit_len_to_buf(RSA_encrypted_premaster_secret)
    return premaster_secret, client_key_exchange

def send_client_FINISHED():
    # https://datatracker.ietf.org/doc/html/rfc5246#section-7.4.9
    # This is not MAC!
    verify_data=prf(master_secret, b"client finished", handshake_hashes_clnt_bin, cryptography.hazmat.primitives.hashes.SHA256(), 12)
    if verbosity>=2:
        print ("verify_data:")
        hexdump.hexdump (verify_data)

    client_finished_pkt=struct.pack(">BB", SSL3_MT_FINISHED, 0)+prepend_16_bit_len_to_buf(verify_data)
    encrypt_MAC_send(_type=SSL3_RT_HANDSHAKE, buf=client_finished_pkt)
    return client_finished_pkt

def send_ClientHello():
    # Random block (32 bytes)
    #ClientHelloRandom=b"\x11"*0x20 # some determinism for debugging
    ClientHelloRandom=os.urandom(0x20)
    if verbosity>=2:
        print ("ClientHelloRandom:")
        hexdump.hexdump(ClientHelloRandom)

    tmp=b""
    tmp=tmp+cur_TLS_ver_bin
    tmp=tmp+ClientHelloRandom
    tmp=tmp+b"\x00"
    tmp=tmp+create_cipher_suites()
    # compression methods length + compression methods:
    tmp=tmp+b"\x01\x00" # this is not length encoded. not a 16-bit value here
    tmp=tmp+create_extensions()

    # the buffer that is hashed in ssl3_finish_mac() (openssl)
    ClientHelloToHash=i2b(SSL3_MT_CLIENT_HELLO) # b"\x01"
    ClientHelloToHash=ClientHelloToHash+b"\x00"+prepend_16_bit_len_to_buf(tmp)

    tmp=i2b(SSL3_RT_HANDSHAKE)+i2w(TLS1_VERSION)+prepend_16_bit_len_to_buf(ClientHelloToHash)
    if verbosity>=2:
        print ("Sending ClientHello (dump without header):")
        hexdump.hexdump(ClientHelloToHash)
    s.send(tmp)
    return ClientHelloRandom, ClientHelloToHash

def unpack_bignum_number(buf):
    return int.from_bytes(buf, byteorder='big')

# https://crypto.stackexchange.com/questions/50886/the-parameters-included-in-tls-1-2-server-signature-in-the-serverkeyexchange
# https://www.rfc-editor.org/rfc/rfc5246#section-4.7
def unpack_SSL3_MT_SERVER_KEY_EXCHANGE(buf):
    assert ciphersuite_after_handshake==TLS_DHE_RSA_WITH_AES_128_CBC_SHA
    global dh_g, dh_p, dh_Ys
    if verbosity>=1:
        print ("unpack_SSL3_MT_SERVER_KEY_EXCHANGE()")
    if verbosity>=2:
        hexdump.hexdump(buf)
    idx=0
    handshake_type=buf[idx]
    assert handshake_type==SSL3_MT_SERVER_KEY_EXCHANGE
    idx=idx+1
    length=(buf[idx]<<16) | (buf[idx+1]<<8) | buf[idx+2]
    if verbosity>=1:
        print ("length", length)
    idx=idx+3
 
    tst=b""

    bignum_len=(buf[idx+0]<<8) | buf[idx+1]
    tst=tst+buf[idx+0:idx+2]
    idx=idx+2
    dh_p=unpack_bignum_number(buf[idx:idx+bignum_len])
    tst=tst+buf[idx:idx+bignum_len]
    idx=idx+bignum_len

    bignum_len=(buf[idx+0]<<8) | buf[idx+1]
    tst=tst+buf[idx+0:idx+2]
    idx=idx+2
    dh_g=unpack_bignum_number(buf[idx:idx+bignum_len])
    tst=tst+buf[idx:idx+bignum_len]
    idx=idx+bignum_len

    bignum_len=(buf[idx+0]<<8) | buf[idx+1]
    tst=tst+buf[idx+0:idx+2]
    idx=idx+2
    dh_Ys=unpack_bignum_number(buf[idx:idx+bignum_len])
    tst=tst+buf[idx:idx+bignum_len]
    idx=idx+bignum_len

    if verbosity>=1:
        print ("dh_p", hex(dh_p))
        print ("dh_g", hex(dh_g))
        print ("dh_Ys AKA pubkey (in Wireshark)", hex(dh_Ys))

    sigalg=struct.unpack(">H", buf[idx:idx+2])[0]
    if verbosity>=1:
        print (f"sigalg {hex(sigalg)}")

    idx=idx+2
    sig_len=(buf[idx+0]<<8) | buf[idx+1]
    idx=idx+2
    sig=buf[idx:idx+sig_len]
    if verbosity>=2:
        print ("sig:")
        hexdump.hexdump(sig)
    idx=idx+sig_len
    try:
        if sigalg==TLSEXT_SIGALG_rsa_pkcs1_sha256:
            RSA_pub_key.verify(sig, ClientHelloRandom + server_random + tst, cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15(), cryptography.hazmat.primitives.hashes.SHA256())
        elif sigalg==TLSEXT_SIGALG_rsa_pkcs1_sha1:
            RSA_pub_key.verify(sig, ClientHelloRandom + server_random + tst, cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15(), cryptography.hazmat.primitives.hashes.SHA1())
        else:
            print (f"Fatal error, unknown sigalg {hex(sigalg)}")
            exit(0)
        """
        NOT TESTED
        elif sigalg==TLSEXT_SIGALG_rsa_pss_pss_sha256:
            RSA_pub_key.verify(sig, ClientHelloRandom + server_random + tst, 
                padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH),
                cryptography.hazmat.primitives.hashes.SHA256())
        """

    except cryptography.exceptions.InvalidSignature:
        print ("Fatal error: can't check RSA signature of DH params (ServerDHParams).")
        print ("Most likely, this is a bug in this code.")
        exit(0)

    assert idx==len(buf) # everything processed

# Get 'server hello' packets
def get_ServerHello_packets_until_SERVER_DONE():
    global MT_SERVER_HELLO_pkt
    global MT_CERTIFICATE_pkt
    global MT_SERVER_DONE_pkt
    global MT_SERVER_KEY_EXCHANGE_pkt
    while True:
        _type, proto_bin, buf=get_data_from_serv()
        proto=struct.unpack(">H", proto_bin)[0]
        if proto!=TLS1_2_VERSION:
            print (f"Serv offers proto {hex(proto)}")
            print ("We currently support only TLS 1.2. Exiting.")
            exit(0)
        handshake_type=buf[0]
        if verbosity>=1:
            print (f"In process of getting ServerHello packets. {_type}")
        if _type==SSL3_RT_ALERT:
            chk_alert(_type, buf)
            exit(0)
        assert _type==SSL3_RT_HANDSHAKE
        if verbosity>=1:
            print (f"Handshake_type {handshake_type}")

        if handshake_type==SSL3_MT_SERVER_HELLO:
            unpack_SSL3_MT_SERVER_HELLO(buf)
            MT_SERVER_HELLO_pkt=buf
        elif handshake_type==SSL3_MT_CERTIFICATE:
            unpack_SSL3_MT_CERTIFICATE(buf)
            MT_CERTIFICATE_pkt=buf
        elif handshake_type==SSL3_MT_SERVER_KEY_EXCHANGE:
            unpack_SSL3_MT_SERVER_KEY_EXCHANGE(buf)
            MT_SERVER_KEY_EXCHANGE_pkt=buf
        elif handshake_type==SSL3_MT_SERVER_DONE:
            unpack_SSL3_MT_SERVER_DONE(buf)
            MT_SERVER_DONE_pkt=buf
            break # last packet in 'server hello' sequence
        else:
            print (f"Error. Unexpected packet from server: handshake_type={handshake_type}")
            exit(0)

def get_serv_packets_until_FINISHED():
    global MT_NEWSESSION_TICKET_pkt
    global RT_CHANGE_CIPHER_SPEC_pkt
    if verbosity>=1:
        print ("Expecting 'new session ticket' message")
        print ("Or 'change cipher spec' message")
        print ("Or 'finished' message")
        print ("Or alert")

    while True:
        try:
            _type, serv_TLS_ver_bin, buf=get_data_from_serv()
        except TimeoutError:
            print ("Timeout during receive. server supposed client should close connection?")
            break
        if verbosity>=1:
            print (f"Got type {_type}")
        if _type==SSL3_RT_ALERT:
            chk_alert(_type, buf)
            exit(0)
        elif _type==SSL3_RT_HANDSHAKE and buf[0]==SSL3_MT_NEWSESSION_TICKET:
            if verbosity>=1:
                print ("Got SSL3_RT_HANDSHAKE SSL3_MT_NEWSESSION_TICKET")
            if verbosity>=2:
                hexdump.hexdump(buf)
            MT_NEWSESSION_TICKET_pkt=buf
        elif _type==SSL3_RT_CHANGE_CIPHER_SPEC:
            if verbosity>=1:
                print ("Got SSL3_RT_CHANGE_CIPHER_SPEC")
            assert buf[0]==1
            RT_CHANGE_CIPHER_SPEC_pkt=buf
        elif _type==SSL3_RT_HANDSHAKE:
            if verbosity>=1:
                print ("Got SSL3_RT_HANDSHAKE")
            # this is the first encrypted/MACed message from server
            tmp=chk_MAC_and_decrypt(_type, serv_TLS_ver_bin, buf)
            assert tmp[0]==SSL3_MT_FINISHED
            if verbosity>=2:
                print ("Decrypted:")
                hexdump.hexdump(tmp)
            chk_server_finished(tmp[4:])
            break # last message in bunch, so break
        else:
            assert False # got unknown packet

def get_all_app_data():
    # Get app data and/or alert (at the end)
    response=b""
    while True:
        try:
            _type, serv_TLS_ver_bin, buf=get_data_from_serv()
        except TimeoutError:
            print ("Timeout during receive. Server supposed client should close the connection?")
            break
        except ValueError:
            break
        if _type==SSL3_RT_ALERT:
            tmp=chk_MAC_and_decrypt(_type, serv_TLS_ver_bin, buf)
            if verbosity>=1:
                print ("Got alert")
            if verbosity>=2:
                hexdump.hexdump(tmp)
            chk_alert(_type, tmp)
            break # we can finish here
        tmp=chk_MAC_and_decrypt(_type, serv_TLS_ver_bin, buf)
        response=response + tmp

    f=open(response_fname, "wb")
    f.write(response)
    f.close()
    print (f"Written to {response_fname} {len(response)} bytes")

def send_close_notify():
    body=struct.pack("BB", SSL3_AL_WARNING, SSL3_AD_CLOSE_NOTIFY)
    encrypt_MAC_send(_type=SSL3_RT_ALERT, buf=body)

def send_HTTP_req():
    # https://www.rfc-editor.org/rfc/rfc7230#section-5.4
    HTTPS_req=b"GET / HTTP/1.1\r\nHost: "+bytes(HOST,"utf-8")+b"\r\nUser-Agent: ToyTLS\r\nAccept: */*\r\n\r\n"
    encrypt_MAC_send(_type=SSL3_RT_APPLICATION_DATA, buf=HTTPS_req)

# global variables

MT_SERVER_HELLO_pkt=None
MT_CERTIFICATE_pkt=None
MT_SERVER_KEY_EXCHANGE_pkt=None # do I need it?
MT_SERVER_DONE_pkt=None
MT_NEWSESSION_TICKET_pkt=None
RT_CHANGE_CIPHER_SPEC_pkt=None

TRY_EXTENDED_MASTER_SECRET=True
#TRY_EXTENDED_MASTER_SECRET=False

cur_seq_n_to_serv=0
cur_seq_n_to_clnt=0

def main():
    global verbosity, the_only_ciphersuite_client_wants, HOST, PORT, DISABLE_ETM
    global s, local_cert_path, ClientHelloToHash, client_key_exchange, premaster_secret
    global handshake_hashes_clnt_bin, master_secret, ClientHelloRandom, client_finished_pkt
    global response_fname, cipher_mode, use_extended_master_secret

    parser = argparse.ArgumentParser()
    parser.add_argument("-host", default="cryptography.re", help="TLS host. Default=cryptography.re")
    parser.add_argument("-port", type=int, default=443, help="TLS port. Default=443")
    parser.add_argument("-disable_etm", default=False, help="Disable ETM mode. Default=false", action="store_true")
    ciphersuites_supported_str=""
    for x in ciphersuites_supported:
        ciphersuites_supported_str=ciphersuites_supported_str+ciphersuite_to_str[x]+" "
    parser.add_argument("-cipher", default="any", help="Ciphersuite. Default: any. Supported: "+ciphersuites_supported_str)
    parser.add_argument("-save", default="output.bin", help="Save server's output to file. Default=output.bin")
    parser.add_argument("-v", action="count", help="Increase verbosity. May be -v, -vv. Basic info, function calls: -v. Hex dumps: -vv")
    parser.add_argument("-local-cert-path", default="/etc/ssl/certs", help="Local certs store. Default: /etc/ssl/certs (as in Ubuntu).")
    
    args=parser.parse_args()

    # https://stackoverflow.com/questions/4042452/display-help-message-with-python-argparse-when-script-is-called-without-any-argu
    parser.print_help()
    print ("")

    DISABLE_ETM=args.disable_etm
    response_fname=args.save
    the_only_ciphersuite_client_wants=args.cipher
    
    if args.v!=None:
        verbosity=args.v
    else:
        verbosity=0

    local_cert_path=args.local_cert_path

    cipher_mode="MTE" # default. older TLS versions
    use_extended_master_secret=False

    HOST=args.host
    PORT=args.port

    s=socket.socket(socket.AF_INET)
    s.connect((HOST, PORT))

    ClientHelloRandom, ClientHelloToHash = send_ClientHello()

    try:
        get_ServerHello_packets_until_SERVER_DONE()
    except TimeoutError:
        print ("TimeoutError during get_ServerHello_packets_until_SERVER_DONE()")
        print ("Can't proceed, exiting")
        exit(0)

    premaster_secret, client_key_exchange=generate_premaster_and_client_key_exchange()

    handshake_hashes_clnt_bin, handshake_hashes_clnt_hex=make_hashes_for_client()
    if verbosity>=2:
        print (f"Handshake_hashes {handshake_hashes_clnt_hex}")

    master_secret=generate_master_secret()
    generate_key_block()

    # client key exchange
    if verbosity>=1:
        print ("sending client key exchange")
    if verbosity>=2:
        print ("client_key_exchange:")
        hexdump.hexdump(client_key_exchange)

    s.send(i2b(SSL3_RT_HANDSHAKE)+cur_TLS_ver_bin+prepend_16_bit_len_to_buf(client_key_exchange))

    # change cipher spec:
    if verbosity>=1:
        print ("sending change cipher spec")
    s.send(i2b(SSL3_RT_CHANGE_CIPHER_SPEC)+cur_TLS_ver_bin+b"\x00\x01\x01")

    client_finished_pkt=send_client_FINISHED()

    get_serv_packets_until_FINISHED()

    send_HTTP_req()

    get_all_app_data()

    # polite client should send the 'close notify' as well
    send_close_notify()

main()

